package net.lemnik.eodsql;

import java.lang.reflect.InvocationTargetException;

import java.sql.Connection;

import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.sql.DataSource;

import net.lemnik.eodsql.impl.DefaultQueryFactory;
import net.lemnik.eodsql.spi.QueryFactory;

/**
 * Created on 2008/06/15
 * @author Jason Morris
 */
class QueryFactoryChain {
    private Link first = null;

    private ThreadLocal<Link> chain = new ThreadLocal<Link>() {
        @Override
        protected Link initialValue() {
            return first;
        }

    };

    private ReentrantReadWriteLock lock = new ReentrantReadWriteLock(false);

    private Lock read = lock.readLock();

    private Lock write = lock.writeLock();

    QueryFactoryChain() {
        add(new DefaultQueryFactory());

        try {
            @SuppressWarnings("unchecked")
            Iterable<QueryFactory> loader = (Iterable<QueryFactory>)Class.forName("java.util.ServiceLoader").getMethod("load", new Class[]{
                        Class.class
                    }).invoke(null, new Object[]{QueryFactory.class});

            for(QueryFactory factory : loader) {
                add(factory);
            }
        } catch(NoClassDefFoundError e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(IllegalArgumentException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(SecurityException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(IllegalAccessException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(InvocationTargetException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(NoSuchMethodException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(ClassNotFoundException e) {
            // ignore this... we are simply not running 1.6 or higher
        }
    }

    <T extends BaseQuery> T create(DataSource ds, Class<T> query, ClassLoader loader) throws InvalidQueryException {
        read.lock();

        T queryImpl = null;

        try {
            Link link = chain.get();

            while(link != null) {
                if(link.getFactory().canConstruct(query)) {
                    queryImpl = link.getFactory().construct(ds, query, loader);
                    break;
                }
            }

            chain.set(link.getNext());
        } finally {
            read.unlock();

            if(lock.getReadLockCount() == 0) {
                chain.remove();
            }
        }

        return queryImpl;
    }

    <T extends BaseQuery> T create(Connection connection, Class<T> query, ClassLoader loader) throws InvalidQueryException {
        read.lock();

        T queryImpl = null;

        try {
            Link link = chain.get();

            while(link != null) {
                if(link.getFactory().canConstruct(query)) {
                    queryImpl = link.getFactory().construct(connection, query, loader);
                    break;
                }
            }

            chain.set(link.getNext());
        } finally {
            read.unlock();

            if(lock.getReadLockCount() == 0) {
                chain.remove();
            }
        }

        return queryImpl;
    }

    void add(QueryFactory factory) {
        write.lock();

        try {
            new Link(factory);
        } finally {
            write.unlock();
        }
    }

    void remove(QueryFactory factory) {
        write.lock();

        try {
            Link link = first;

            while(link != null && link.getNext() != null) {
                if(link.getFactory() == factory) {
                    link.prev.next = link.next;
                    link.next.prev = link.prev;
                    break;
                }
            }
        } finally {
            write.unlock();
        }
    }

    private class Link {
        private QueryFactory factory;

        private Link next = null;

        private Link prev = null;

        Link(QueryFactory factory) {
            this.factory = factory;

            if(first != null) {
                first.prev = this;
            }

            this.next = first;
            first = this;
        }

        Link getNext() {
            return next;
        }

        Link getPrev() {
            return prev;
        }

        QueryFactory getFactory() {
            return factory;
        }

        @Override
        public String toString() {
                return "FactoryLink[" + factory.getClass().getName() + "]";
        }

    }
}
